package mushroom;

import java.util.*;

/**
* B-Mushroom Interpreter 1.0
* An open-source Befunge interpreter.
*/
public class Interpreter implements Runnable
{
    Stack<Short> mainStack = new Stack<Short>();
    Stack<Short> stack1 = new Stack<Short>();
    Stack<Short> stack2 = new Stack<Short>();
    Stack<Short> stack3 = new Stack<Short>();
    Stack<Short> stack4 = new Stack<Short>();
    //I doubt more than 4 new stacks are needed...
    Stack<Short> currentStack = mainStack;

    short[][] matrix;
    
    String code;
    
    public Interpreter(String code)
    {
        this.code = code;
        System.out.println("B-Mushroom Interpreter 1.0");
        new Thread(this).start();
    }
    public void run()
    {
        int size = (int) Math.sqrt(code.length())* 3 / 2;
        matrix = new short[size][size];//this might not be the most memory-efficient solution...
        //now load everything into the matrix
        int xpos=0,ypos=0;
        for(int i=0;i<code.length();i++)
        {
            char character = code.charAt(i);
            if(character=='\n')
            {
                xpos=0;
                ypos++;
                continue;
            }
            matrix[xpos][ypos] = (short)character;
            //the matrix has you now, character
            xpos++;
        }
        //loading is done.
        xpos=0;//remember these from line 28?
        ypos=0;
        short a,b,x,y;
        boolean isStrMode=false;
        boolean skipNextCell=false;
        Direction currdir = Direction.RIGHT;
        //begin the program loop

        mainloop:
        while(true)
        {
            if(!isStrMode)
            {
                switch( (char)matrix[xpos][ypos] )
                {
                    case '>':
                        currdir = Direction.RIGHT;
                        break;
                    case '<':
                        currdir = Direction.LEFT;
                        break;
                    case '^':
                        currdir = Direction.UP;
                        break;
                    case 'v':
                        currdir = Direction.DOWN;
                        break;
                    case '@':
                        break mainloop;
                    case '0':
                        currentStack.push((short)0);
                        break;
                    case '1':
                        currentStack.push((short)1);
                        break;
                    case '2':
                        currentStack.push((short)2);
                        break;
                    case '3':
                        currentStack.push((short)3);
                        break;
                    case '4':
                        currentStack.push((short)4);
                        break;
                    case '5':
                        currentStack.push((short)5);
                        break;
                    case '6':
                        currentStack.push((short)6);
                        break;
                    case '7':
                        currentStack.push((short)7);
                        break;
                    case '8':
                        currentStack.push((short)8);
                        break;
                    case '9':
                        currentStack.push((short)9);
                        break;
                    case '+':
                        a = currentStack.pop();
                        b = currentStack.pop();
                        currentStack.push((short)(b+a));
                        break;
                    case '-':
                        //b-a. This is important!
                        a = currentStack.pop();
                        b = currentStack.pop();
                        currentStack.push((short)(b-a));
                        break;
                    case '*':
                        a = currentStack.pop();
                        b = currentStack.pop();
                        currentStack.push((short)(b*a));
                        break;
                    case '/':
                        //b/a. This is important!
                        a = currentStack.pop();
                        b = currentStack.pop();
                        currentStack.push((short)(b/a));
                        break;
                    case '%':
                        a = currentStack.pop();
                        b = currentStack.pop();
                        currentStack.push((short)(b%a));
                        break;
                    case '!':
                        a = currentStack.pop();
                        if(a==0) currentStack.push((short)1);
                        else currentStack.push((short)0);
                        break;
                    case '`':
                        a = currentStack.pop();
                        b = currentStack.pop();
                        if(b>a) currentStack.push((short)1);
                        else currentStack.push((short)0);
                        break;
                    case '"':
                        isStrMode = !isStrMode;
                        break;
                    case '?':
                        int randomDir = (int) Math.floor( 0.25*4 );
                        if(randomDir==0) currdir = Direction.UP;
                        else if(randomDir==1) currdir = Direction.RIGHT;
                        else if(randomDir==2) currdir = Direction.DOWN;
                        else currdir = Direction.LEFT;
                        break;
                    case '_':
                        a = currentStack.pop();
                        if(a==0) currdir = Direction.RIGHT;
                        else currdir = Direction.LEFT;
                        break;
                    case '|':
                        a = currentStack.pop();
                        if(a==0) currdir = Direction.DOWN;
                        else currdir = Direction.RIGHT;
                        break;
                    case ':':
                        a = currentStack.peek();
                        currentStack.push(a);
                        break;
                    case '\\':
                        a = currentStack.pop();
                        b = currentStack.pop();
                        currentStack.push(a);
                        currentStack.push(b);
                        break;
                    case '$':
                        currentStack.pop();//get rid of the value.
                        break;
                    case '.':
                        a = currentStack.pop();
                        System.out.println(a);
                        break;
                    case ',':
                        a = currentStack.pop();
                        System.out.println( (char)a );
                        break;
                    case '#':
                        skipNextCell=true;
                        break;
                    case 'p':
                        x = currentStack.pop();
                        y = currentStack.pop();
                        byte v = currentStack.pop().byteValue();
                        matrix[x][y]=v;
                        break;
                    case 'g':
                        x = currentStack.pop();
                        y = currentStack.pop();
                        currentStack.push(matrix[x][y]);
                        break;
                    case '&':
                        Scanner sc = new Scanner(System.in);
                        a = sc.nextByte();
                        currentStack.push(a);
                        break;
                    case '~':
                        Scanner sc2 = new Scanner(System.in);
                        String input = sc2.nextLine();
                        char a_ = input.charAt(0);
                        currentStack.push( (short)a_ );
                        break;
                    case ';':
                        a = currentStack.pop();
                        if(a==0) currentStack = mainStack;
                        else
                        {
                            switch(a)//this isn't the best solution, but it's the easiest one that works
                            {
                                case 1:
                                    currentStack = stack1;
                                    break;
                                case 2:
                                    currentStack = stack2;
                                    break;
                                case 3:
                                    currentStack = stack3;
                                    break;
                                case 4:
                                    currentStack = stack4;
                                    break;
                                default:
                                    currentStack = mainStack;
                                    break;
                            }
                        }
                }
            }
            else if(isStrMode)
            {
                if( (char)matrix[xpos][ypos] == '"') isStrMode=false;
                else mainStack.push( matrix[xpos][ypos] );
            }
            else if(skipNextCell)
            {
                skipNextCell=false;
            }
            switch(currdir)
            {
                case UP:
                    if(ypos==0) ypos=matrix[0].length-1;
                    else ypos--;
                    break;
                case RIGHT:
                    if(xpos==matrix.length) xpos=0;
                    else xpos++;
                    break;
                case DOWN:
                    if(ypos==matrix[0].length) ypos=0;
                    else ypos++;
                    break;
                case LEFT:
                    if(xpos==0) xpos=matrix.length;
                    else xpos--;
                    break;
            }
            //end entire loop.
        }//end while
    }
}
