package pku.sei.dependencegraph;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.lang.management.ManagementFactory;
import java.lang.management.MemoryMXBean;
import java.lang.management.MemoryUsage;
import java.util.HashMap;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeSet;

import pku.sei.dependencegraph.algorithm.App_CFL_RHS;
import pku.sei.dependencegraph.algorithm.App_TAL_RHS;
import pku.sei.dependencegraph.algorithm.CFL_RHS;
import pku.sei.dependencegraph.algorithm.Lib_CFL_RHS;
import pku.sei.dependencegraph.algorithm.Lib_TAL_RHS;
import pku.sei.dependencegraph.algorithm.summary.CflSummary;
import pku.sei.dependencegraph.algorithm.summary.TalSummary;
import pku.sei.dependencegraph.graph.Graph;
import pku.sei.dependencegraph.graph.INodeType;

public class Main {
	public static void main(String[] args) {;
		//PrintStream ps = null;
		//try {
			//ps = new PrintStream(new File("Cache.txt"));
		
			Graph dg = new DependenceGraphExtractor().run(args[0], args[1], args[2], args[3]);
			dg.propagateCallInfoAndLabelKeyNodes();
			dg.labelBoundary();			
			display(dg);
				
			
			long starttime, endtime;
			long[] latexform = new long[8];
			
			MemoryMXBean memoryMXBean = ManagementFactory.getMemoryMXBean();
			starttime = System.currentTimeMillis();
			CflSummary cflSum = new Lib_CFL_RHS(dg).cflAlgorithmForLib();
			endtime = System.currentTimeMillis();
			System.out.println("Lib_CFL time = " + (endtime - starttime));
			System.out.println("        Used = " + memoryMXBean.getHeapMemoryUsage().getUsed() / (1 << 20));
			latexform[4] = endtime - starttime;
			latexform[5] = memoryMXBean.getHeapMemoryUsage().getUsed() / (1 << 20);
			
			System.gc();
			starttime = System.currentTimeMillis();
			new App_CFL_RHS(dg, cflSum).cflAlgorithmAddApp();
			endtime = System.currentTimeMillis();
			System.out.println("App_CFL time = " + (endtime - starttime));	
			System.out.println("        Used = " + memoryMXBean.getHeapMemoryUsage().getUsed() / (1 << 20));	
			latexform[0] = endtime - starttime;
			latexform[1] = memoryMXBean.getHeapMemoryUsage().getUsed() / (1 << 20);
						
			starttime = System.currentTimeMillis();
			TalSummary ts = new Lib_TAL_RHS(dg).talAlgorithmForLib();
			endtime = System.currentTimeMillis();
			System.out.println("Lib_TAL time = " + (endtime - starttime));
			System.out.println("        Used = " + memoryMXBean.getHeapMemoryUsage().getUsed() / (1 << 20));
			latexform[6] = endtime - starttime;
			latexform[7] = memoryMXBean.getHeapMemoryUsage().getUsed() / (1 << 20);
			
			System.gc();
			starttime = System.currentTimeMillis();
			new App_TAL_RHS(dg, ts).cflAlgorithmAddApp();
			endtime = System.currentTimeMillis();
			System.out.println("App_TAL time = " + (endtime - starttime));
			System.out.println("        Used = " + memoryMXBean.getHeapMemoryUsage().getUsed() / (1 << 20));
			latexform[2] = endtime - starttime;
			latexform[3] = memoryMXBean.getHeapMemoryUsage().getUsed() / (1 << 20);
			System.out.println(java.util.Arrays.toString(latexform));
			
			//ps.close();
		//} catch (FileNotFoundException e) {
		//	e.printStackTrace();
		//}
	}
	private static void display(Graph dg) {
		int n = dg.getN();
		
		int clientSize = 0;
		for (int i = 0; i < n; i++) {
			if (dg.getNodeIsClientCode(i)) {
				clientSize ++;
//				dg.initializeForwardEdgeIterator(i);				
//				while (dg.hasNextEdge()) {
//					int tgt = dg.getForwardEdgeTarget();
//					if (!dg.getNodeIsClientCode(tgt)) {
//						System.out.println("APP[" + i + "]->LIB[" + tgt + "]");
//						System.out.println("	APP:" + dg.getStatement(i));
//						System.out.println("	LIB:" + dg.getStatement(tgt));
//					}
//				}
//				dg.initializeBackwardEdgeIterator(i);
//				while (dg.hasNextEdge()) {
//					int src = dg.getBackwardEdgeSource();
//					if (!dg.getNodeIsClientCode(src) && dg.getNodeType(src) == INodeType.CALL_NODE) { 
//						System.out.println("APP[" + i + "]<-LIB[" + src + "]");
//						System.out.println("	APP:" + dg.getStatement(i));
//						System.out.println("	LIB:" + dg.getStatement(src));
//					}
//				}			
			}
		}
		System.out.println("|client|=" + clientSize + ";" + "|Graph|=" + n + "\n");
		int CallCount, ReturnCount, EntryCount, ExitCount;
		CallCount = ReturnCount = EntryCount = ExitCount = 0;
		int ReturnHasCallInfo = 0;
		Set<Integer> keyMethodSet = new TreeSet<Integer>();
		for (int i = 0; i < n; i++) {
			switch (dg.getNodeType(i)) {
			case INodeType.CALL_NODE:
				CallCount++;
				break;
			case INodeType.RETURN_NODE:
				ReturnCount++;
				if (dg.getNodeHasCallInfo(i)) {
					ReturnHasCallInfo++;
					keyMethodSet.add(dg.getMethodId(i));
				}
				break;
			case INodeType.ENTRY_NODE:
				EntryCount++;
				break;
			case INodeType.EXIT_NODE:
				ExitCount++;
				break;
			}
		}
		/*
		System.out.println("Total:" + n);
		System.out.println("	Call:" + CallCount);
		System.out.println("	Return:" + ReturnCount);
		System.out.println("		Return+C:" + ReturnHasCallInfo);
		System.out.println("	Entry:" + EntryCount);
		System.out.println("	Exit:" + ExitCount);
		System.out.println("KeyMethod:" + keyMethodSet.size());
		*/
		int keyEntry, keyExit;
		keyEntry = keyExit = 0;
		for (int i = 0; i < n; i++) {
			if (dg.getNodeType(i) == INodeType.ENTRY_NODE && keyMethodSet.contains(dg.getMethodId(i))) {
				keyEntry++;
			}
			else if (dg.getNodeType(i) == INodeType.EXIT_NODE && keyMethodSet.contains(dg.getMethodId(i))) {
				keyExit++;
			}
		}
		/*
		System.out.println("	KeyEntry:" + keyEntry);
		System.out.println("	KeyExit:" + keyExit);
		*/
		int totalKeyNode = 0;
		for (int i = 0; i < n; i++)
			if (dg.getNodeIsKey(i))
				totalKeyNode++;
		System.out.println("Total Fundamental Node:" + totalKeyNode + "\n");
		
		HashMap<String, Integer> methodCount = new HashMap<String, Integer>();
		
		for (int i = 0; i < n; i++) {
			String mi = dg.getStatement(i).getNode().getMethod().toString();
			if (!methodCount.containsKey(mi)) {
				methodCount.put(mi, 0);
			}
			methodCount.put(mi, methodCount.get(mi)+1);
		}
	}
	
	
}
