package pku.sei.dependencegraph;

import java.util.Iterator;
import java.util.Set;
import java.util.HashSet;
import com.ibm.wala.classLoader.IClass;
import com.ibm.wala.classLoader.IField;
import com.ibm.wala.classLoader.IMethod;
import com.ibm.wala.classLoader.IBytecodeMethod;
import com.ibm.wala.classLoader.CallSiteReference;
import com.ibm.wala.ipa.cha.IClassHierarchy;
import com.ibm.wala.ssa.IR;
import com.ibm.wala.ssa.SSAInstruction;
import com.ibm.wala.ssa.analysis.IExplodedBasicBlock;
import com.ibm.wala.types.TypeName;
import com.ibm.wala.types.TypeReference;
import com.ibm.wala.types.MethodReference;
import com.ibm.wala.util.graph.Graph;
import com.ibm.wala.ipa.callgraph.CallGraph;
import com.ibm.wala.ipa.callgraph.CGNode;
import com.ibm.wala.ipa.cfg.BasicBlockInContext;


public class WALAUtils
{
	public static String getWALAClassName(String className)
	{
		return "L"+className.replace('.', '/');	
	}
	
	public static String getPackageName(String fullClassName)
    {
		String packageName = "";
		int k = fullClassName.lastIndexOf('.');
		if (k != -1) packageName = fullClassName.substring(0, k);
		return packageName;
	}

	public static String getClassName(TypeReference typeRef) throws Exception
	{
		return getTypeNameStr(typeRef);
	}
	
	public static String getClassName(IClass cls) throws Exception
	{
		if (cls == null) return null;
		
		String cnStr = getTypeNameStr(cls.getName());
		
		return cnStr;
	}
	
	public static String getFieldName(IField field) throws Exception
	{
		if (field == null) return null;
		
		TypeName classTypeName = field.getDeclaringClass().getName();
		
		StringBuffer fnStrBuf = new StringBuffer();

		fnStrBuf.append(getTypeNameStr(classTypeName));
		fnStrBuf.append(".");
		fnStrBuf.append(field.getName().toUnicodeString());
		
		return fnStrBuf.toString();
	}
	
	public static String getDeclaringClassNameForField(IField field) throws Exception
	{
		if (field == null) return null;
		
		TypeName classType = field.getDeclaringClass().getName();
		
		return getTypeNameStr(classType);
	}
	
	public static String getFieldTypeClassName(IField field) throws Exception
	{
		if (field == null) return null;
		
		TypeReference fieldTypeRef = field.getFieldTypeReference();
		if ( ! fieldTypeRef.isReferenceType() ) return null;

		TypeName fieldTypeClass = fieldTypeRef.getName();

		return getTypeNameStr(fieldTypeClass);
	}
	
	public static String getTypeNameStr(TypeName typeName) throws Exception
	{
		StringBuffer typeStrBuf = new StringBuffer();
			
		if (typeName.getPackage() != null) typeStrBuf.append(typeName.getPackage().toUnicodeString().replace('/', '.')).append(".");

		typeStrBuf.append(typeName.getClassName().toUnicodeString());
			
		if (typeName.isArrayType()) typeStrBuf.append("[]");
			
		String typeStr = typeStrBuf.toString();

		return typeStr;
	}

	public static String getTypeNameStr(TypeReference typeRef)
	{
		if (typeRef == null) return null;
		
		if ( ! typeRef.isReferenceType() ) return null;
		
		try
		{
			TypeName typeClass = typeRef.getName();
			
			String typeStr = getTypeNameStr(typeClass);

			return typeStr;
		}
		catch (Exception ex) { ex.printStackTrace(); }
		
		return null;
	}
	
	public static String getMethodSignature(IMethod mth)
	{
		return mth.getSignature();
	}

	public static String getMethodSignature(MethodReference mthRef)
	{
		return mthRef.getSignature();
	}	

	public static String getShortMethodName(MethodReference mthRef) throws Exception
	{
		return mthRef.getName().toUnicodeString();
	}
	
	public static String getFullMethodNameFromCGNode(CGNode node) throws Exception
	{	
		IClass cls = node.getMethod().getDeclaringClass();
		
		String className = getClassName(cls);
		
		String methodName = node.getMethod().getReference().getName().toUnicodeString();
				
		StringBuffer mnStrBuf = new StringBuffer();

		mnStrBuf.append(className).append(".").append(methodName);

		return mnStrBuf.toString();
	}
	
	public static String getClassNameFromCGNode(CGNode node) throws Exception
	{	
		IClass cls = node.getMethod().getDeclaringClass();
		
		return getClassName(cls);
	}
	
	public static String getMethodNameFromCGNode(CGNode node) throws Exception
	{	
		return node.getMethod().getReference().getName().toUnicodeString();
	}
	
	public static String getDeclaringClassNameForMethod(IMethod mth) throws Exception
	{
		IClass cls = mth.getDeclaringClass();
		
		return getClassName(cls);
	}
	
	public static boolean isMethodConstructor(CGNode node) throws Exception
	{	
		return node.getMethod().isInit();
	}
	
	public static IClass findClass(IClassHierarchy cha, String className) throws Exception
	{
		if (className.endsWith("[]")) return null;
		
		Iterator<IClass> clsIt = cha.iterator();
		while (clsIt.hasNext())
		{
			IClass cls = clsIt.next();
			
			if (className.equals(getClassName(cls))) return cls;
		}
		
		return null;
	}
	
	public static int getInsnBytecodeIndex(CGNode mthNode, int ssaInsnIndex)
	{
		return getInsnBytecodeIndex(mthNode.getMethod(), ssaInsnIndex);
	}
	
	public static int getInsnBytecodeIndex(IMethod mth, int ssaInsnIndex)
	{
		try
		{
			return ((IBytecodeMethod) mth).getBytecodeIndex(ssaInsnIndex);
		}
		catch (Exception ex)
		{
			return -1;
		}
	}
	
	public static void printCallGraph(CallGraph clGraph, int maxLevel)
	{
		System.out.println("CALL GRAPH");
		System.out.println("==========");

		CGNode entryNode = clGraph.getFakeRootNode(); 
		printCallGraphNode(clGraph, entryNode, "method: ", maxLevel, 0);

		System.out.println("");
	}

	private static void printCallGraphNode(CallGraph clGraph, CGNode node, String prefix, int maxLevel, int curLevel)
	{
		if (curLevel > maxLevel) return;
		
		System.out.println(prefix + getMethodSignature(node.getMethod()));
		
		Iterator<CallSiteReference> callSitesIt = node.iterateCallSites();
		while (callSitesIt.hasNext())
		{
			CallSiteReference callSite = callSitesIt.next();
			
			Set<CGNode> targetNodes = clGraph.getPossibleTargets(node, callSite);
			
			for (CGNode tgtNode : targetNodes) printCallGraphNode(clGraph, tgtNode, "\t" + prefix, maxLevel, curLevel + 1);
		}
	}

	public static void printMethodIR(CGNode mthNode, String mthSig)
	{
		IR methodIR = mthNode.getIR();
					
		if (methodIR == null) return;
					
		System.out.println("method signature = " + mthSig);
	   
		SSAInstruction[] instructions = methodIR.getInstructions();
		for (int ssaIndex = 0; ssaIndex < instructions.length; ssaIndex++)
		{
			int insnPos = WALAUtils.getInsnBytecodeIndex(mthNode, ssaIndex);
			if (instructions[ssaIndex] == null) System.out.println("\t " + insnPos + ": null");
			else System.out.println("\t " + insnPos + ": " + instructions[ssaIndex].toString());
		}
	}
	
	public static void printAllMethodsIR(Graph<BasicBlockInContext<IExplodedBasicBlock>> icfg)
	{
		Set<String> printedMethods = new HashSet<String>();
				
		System.out.println("METHOD SSA IR");
		System.out.println("============="); 
		
		for (BasicBlockInContext<IExplodedBasicBlock> bb : icfg) 
		{
			CGNode node = bb.getNode();
					
			String methodSig = getMethodSignature(node.getMethod());
			
			if (printedMethods.contains(methodSig)) continue;
			
			printMethodIR(node, methodSig);

			printedMethods.add(methodSig);
		}

		System.out.println("");
	}

	public static void printAllMethodsIR(CallGraph clGraph)
	{
		Set<String> printedMethods = new HashSet<String>();
				
		System.out.println("METHOD SSA IR");
		System.out.println("============="); 
		
		for (CGNode node : clGraph) 
		{
			String methodSig = getMethodSignature(node.getMethod());
			
			if (printedMethods.contains(methodSig)) continue;
			
			printMethodIR(node, methodSig);

			printedMethods.add(methodSig);
		}

		System.out.println("");
	}
	
	public static void printInterProcCFG(Graph<BasicBlockInContext<IExplodedBasicBlock>> icfg)
	{
		System.out.println("INTERPROC CFG");
		System.out.println("============="); 
		
		for (BasicBlockInContext<IExplodedBasicBlock> bb : icfg) 
		{
			for (Iterator<BasicBlockInContext<IExplodedBasicBlock>> succIt = icfg.getSuccNodes(bb); succIt.hasNext(); )
			{
				BasicBlockInContext<IExplodedBasicBlock> succ = succIt.next();
			
				String methodSigFrom = getMethodSignature(bb.getNode().getMethod());
				String methodSigTo = getMethodSignature(succ.getNode().getMethod());
				
				System.out.println(bb.toString() + " ==> " + succ.toString());
			}
		}
			
		System.out.println("");
	}

}

