package pku.sei.dependencegraph.algorithm;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Queue;
import java.util.Set;

import pku.sei.dependencegraph.algorithm.mapkey.PairPtypeNode;
import pku.sei.dependencegraph.graph.Graph;
import pku.sei.dependencegraph.graph.INodeType;
import pku.sei.dependencegraph.graph.IPathType_1;
import pku.sei.dependencegraph.graph.Path_1;

public class CFL_RHS {
	private Graph g;
	public CFL_RHS(Graph g) {
		this.g = g;
	}
	private Queue<Path_1> w1 = new LinkedList<Path_1>();  ; // 1-order worklist
	private Set<Path_1> r1 = new HashSet<Path_1>(); // 1-order result
	
	private HashMap<PairPtypeNode, ArrayList<Path_1>> srcCandPathsMap
		= new HashMap<PairPtypeNode, ArrayList<Path_1>>();
	private HashMap<PairPtypeNode, ArrayList<Path_1>> tgtCandPathsMap
		= new HashMap<PairPtypeNode, ArrayList<Path_1>>();
	
	/**
	 * 
	 * @param candPathsMap
	 * @param pathType
	 * @param pivotNode intersection{e1,e2}=pivotNode  =>  e1,e2->concatenate(e1,e2)
	 */
	private void pushToIndexes(
			HashMap<PairPtypeNode, ArrayList<Path_1>> candPathsMap, 
			int pathType, int pivotNode, Path_1 path) {
		PairPtypeNode key = new PairPtypeNode(pathType, pivotNode);
		if (!candPathsMap.containsKey(key))
			candPathsMap.put(key, new ArrayList<Path_1>());
		ArrayList<Path_1> pathList = candPathsMap.get(key);
		pathList.add(path);
	}
	
	private ArrayList<Path_1> getCandidatePaths(
			HashMap<PairPtypeNode, ArrayList<Path_1>> candPathsMap,
			int pathType, int pivotNode) {
		PairPtypeNode key = new PairPtypeNode(pathType, pivotNode);
		if (!candPathsMap.containsKey(key))
			return new ArrayList<Path_1>();
		return candPathsMap.get(key);
	}
	
	private void propagateProduction(Path_1 infer) {
		if (!r1.contains(infer)) {
			r1.add(infer);
			if (IPathType_1.toBeExpanded(infer.pathType)) {
				w1.offer(infer);
				pushToIndexes(srcCandPathsMap, infer.pathType, infer.src, infer);
				pushToIndexes(tgtCandPathsMap, infer.pathType, infer.tgt, infer);
			}
		}
	}
	
	/*
	 * A -> @B C 
	 */
	private void propagateRightAppend(int TypeA, Path_1 pB, int TypeC) {
		ArrayList<Path_1> candPaths 
			= getCandidatePaths(srcCandPathsMap, TypeC, pB.tgt);
		for (Path_1 pC : candPaths) {
			Path_1 pA = new Path_1(pB.src, pC.tgt, TypeA);
			propagateProduction(pA);
			//System.out.println("A -> @B C ");
		}
	}
	/*
	 * A -> B @C
	 */
	private void propagateLeftAppend(int TypeA, int TypeB, Path_1 pC) {
		ArrayList<Path_1> candPaths 
			= getCandidatePaths(tgtCandPathsMap, TypeB, pC.src);
		for (Path_1 pB : candPaths) {
			Path_1 pA = new Path_1(pB.src, pC.tgt, TypeA);
			propagateProduction(pA);
			//System.out.println("A -> B @C");
		}
	}	
	/*
	 * A -> @a
	 */
	private void propagateTrans(int TypeA, Path_1 pa) {
		Path_1 pA = new Path_1(pa.src, pa.tgt, TypeA);
		propagateProduction(pA);
		//System.out.println("A -> @a");
	}
	/*
	 * S -> {i@Si}
	 */
	private void propagateBraces(Path_1 p) {
		if (g.getNodeType(p.src) == INodeType.ENTRY_NODE && g.getNodeType(p.tgt) == INodeType.EXIT_NODE
				&& g.getMethodId(p.src) == g.getMethodId(p.tgt)) {
			// collect possibly appearing brace's path type
			g.initializeBackwardEdgeIterator(p.src);
			HashSet<Integer> braceSet = new HashSet<Integer>();
			while (g.hasNextEdge()) {
				int brace = g.getEdgeType();
				if (!braceSet.contains(brace)) {
					braceSet.add(brace);
				}
			}
			Iterator<Integer> braceIt = braceSet.iterator();
			while (braceIt.hasNext()) {
				Integer bID = braceIt.next();
				ArrayList<Path_1> candLeftBrace 
					= getCandidatePaths(tgtCandPathsMap, bID, p.src);
				ArrayList<Path_1> candRightBrace 
					= getCandidatePaths(srcCandPathsMap, -bID, p.tgt);
				for (Path_1 eLeft: candLeftBrace)
					for (Path_1 eRight: candRightBrace) {
						// eLeft -> p -> eRight
						Path_1 pS = new Path_1(eLeft.src, eRight.tgt, IPathType_1.S_S);
						propagateProduction(pS);
						//System.out.println("S -> {i@Si}");
					}
			}
			
		}
	}
	
	public void cflAlgorithm() {
		/*
		 * 'e' : labelled 0
		 * '{_i' : labelled i
		 * '}_i' : labelled -i
		 */
		// Init the worklist
		int n = g.getN();
		for (int srcId = 0; srcId < n; srcId++) {
			g.initializeForwardEdgeIterator(srcId);
			while (g.hasNextEdge()) {
				int tgtId = g.getForwardEdgeTarget();
				int pathType = g.getEdgeType();
				propagateProduction(new Path_1(srcId, tgtId, pathType));
			}
		}
		System.gc();
		// Add Reachability for production
		int cnt = 0;
		while (!w1.isEmpty()) {
			Path_1 p = w1.poll();
			cnt ++;
			if (cnt % 10000 == 0)
				System.out.println(r1.size() + " & " + w1.size());
			switch (p.pathType) {
				case IPathType_1.S_S:
					// S -> @SS
					propagateRightAppend(IPathType_1.S_S, p, IPathType_1.S_S);
					// S -> S@S
					propagateLeftAppend(IPathType_1.S_S, IPathType_1.S_S, p);
					// S -> {i@Si}
					propagateBraces(p);
					// L -> L@S
					propagateLeftAppend(IPathType_1.S_SL, IPathType_1.S_SL, p);
					// L -> @SL
					propagateRightAppend(IPathType_1.S_SL, p, IPathType_1.S_SL);
					// R -> R@S
					propagateLeftAppend(IPathType_1.S_SR, IPathType_1.S_SR, p);
					// R -> @SR
					propagateRightAppend(IPathType_1.S_SR, p, IPathType_1.S_SR);
					break;
				case IPathType_1.S_SL:
					// L -> @LS
					propagateRightAppend(IPathType_1.S_SL, p, IPathType_1.S_S);
					// L -> S@L
					propagateLeftAppend(IPathType_1.S_SL, IPathType_1.S_S, p);
					// L -> @LL
					propagateRightAppend(IPathType_1.S_SL, p, IPathType_1.S_SL);
					// L -> L@L
					propagateLeftAppend(IPathType_1.S_SL, IPathType_1.S_SL, p);
					// M -> R@L
					propagateLeftAppend(IPathType_1.S_SM, IPathType_1.S_SR, p);
					break;
				case IPathType_1.S_SR:
					// R -> @RS
					propagateRightAppend(IPathType_1.S_SR, p, IPathType_1.S_S);
					// R -> S@R
					propagateLeftAppend(IPathType_1.S_SR, IPathType_1.S_S, p);
					// R -> @RR
					propagateRightAppend(IPathType_1.S_SR, p, IPathType_1.S_SR);
					// R -> R@R
					propagateLeftAppend(IPathType_1.S_SR, IPathType_1.S_SR, p);
					// M -> @RL
					propagateRightAppend(IPathType_1.S_SM, p, IPathType_1.S_SL);
					break;
				case IPathType_1.S_SM: 
					// will not be processed 
					// since it does not appear in the right-side of any production
					break;
				default:
					if (p.pathType > 0)
					// L -> @{i
						propagateTrans(IPathType_1.S_SL, p);
					else 
					// R -> @}i
						propagateTrans(IPathType_1.S_SR, p);
					// {i S }i will be processed in "S_S"
			}
			//System.gc();
		}	
		System.gc();
		
		System.out.println(r1.size());
		
		
	}

}
