package pku.sei.dependencegraph.algorithm;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Queue;
import java.util.Set;

import pku.sei.dependencegraph.algorithm.mapkey.PairNodes;
import pku.sei.dependencegraph.algorithm.mapkey.PairPtypeNode;
import pku.sei.dependencegraph.graph.Graph;
import pku.sei.dependencegraph.graph.INodeType;
import pku.sei.dependencegraph.graph.IPathType_1;
import pku.sei.dependencegraph.graph.IPathType_2;
import pku.sei.dependencegraph.graph.Path_1;
import pku.sei.dependencegraph.graph.Path_2;

public class TAL_RHS {
	private Graph g;
	public TAL_RHS(Graph g) {
		this.g = g;
	}
	private Queue<Path_1> w1 = new LinkedList<Path_1>();  ; // 1-order worklist
	private Set<Path_1> r1 = new HashSet<Path_1>(); // 1-order result
	private Queue<Path_2> w2 = new LinkedList<Path_2>();  ; // 2-order worklist
	private Set<Path_2> r2 = new HashSet<Path_2>(); // 2-order result
	
	private HashMap<PairPtypeNode, ArrayList<Path_1>> srcCandPathsMap
		= new HashMap<PairPtypeNode, ArrayList<Path_1>>();
	private HashMap<PairPtypeNode, ArrayList<Path_1>> destCandPathsMap
		= new HashMap<PairPtypeNode, ArrayList<Path_1>>();
	
	private HashMap<Integer, ArrayList<Path_1>> typeRBraceMap
		= new HashMap<Integer, ArrayList<Path_1>>();
	
	private HashMap<Integer, ArrayList<Path_2>> s1Map
		= new HashMap<Integer, ArrayList<Path_2>>();
	private HashMap<Integer, ArrayList<Path_2>> d1Map
		= new HashMap<Integer, ArrayList<Path_2>>();
	private HashMap<Integer, ArrayList<Path_2>> s2Map
		= new HashMap<Integer, ArrayList<Path_2>>();
	private HashMap<Integer, ArrayList<Path_2>> d2Map
		= new HashMap<Integer, ArrayList<Path_2>>();
	
	private HashMap<PairNodes, ArrayList<Path_2>> innerPairMap
		= new HashMap<PairNodes, ArrayList<Path_2>>();
	private HashMap<PairNodes, ArrayList<Path_2>> outerPairMap
		= new HashMap<PairNodes, ArrayList<Path_2>>();
	
	/**
	 * 
	 * @param candPathsMap
	 * @param pathType
	 * @param pivotNode intersection{e1,e2}=pivotNode  =>  e1,e2->concatenate(e1,e2)
	 */
	private void pushToIndexes(
			HashMap<PairPtypeNode, ArrayList<Path_1>> candPathsMap, 
			int pathType, int pivotNode, Path_1 path) {
		PairPtypeNode key = new PairPtypeNode(pathType, pivotNode);
		if (!candPathsMap.containsKey(key))
			candPathsMap.put(key, new ArrayList<Path_1>());
		ArrayList<Path_1> pathList = candPathsMap.get(key);
		pathList.add(path);
	}
	
	/**
	 * left(right)-inner(outer) append
	 * @param sdMap
	 * @param pivotNode
	 * @param path (intersection = pivotNode)
	 */
	private void pushToSdMap(
			HashMap<Integer, ArrayList<Path_2>> sdMap, 
			int pivotNode, Path_2 dPath) {
		if (!sdMap.containsKey(pivotNode))
			sdMap.put(pivotNode, new ArrayList<Path_2>());
		ArrayList<Path_2> dPathList = sdMap.get(pivotNode);
		dPathList.add(dPath);
	}
	
	/**
	 * 
	 * @param sdMap
	 * @param node1
	 * @param node2
	 */
	private void pushToPairMap(
			HashMap<PairNodes, ArrayList<Path_2>> dMap, 
			int node1, int node2, Path_2 dPath) {
		PairNodes dnodes = new PairNodes(node1, node2); 
		if (!dMap.containsKey(dnodes))
			dMap.put(dnodes, new ArrayList<Path_2>());
		ArrayList<Path_2> dPathList = dMap.get(dnodes);
		dPathList.add(dPath);
	}
	
	private ArrayList<Path_1> getCandidatePaths(
			HashMap<PairPtypeNode, ArrayList<Path_1>> candPathsMap,
			Integer pathType, int pivotNode) {
		PairPtypeNode key = new PairPtypeNode(pathType, pivotNode);
		if (!candPathsMap.containsKey(key))
			return new ArrayList<Path_1>();
		return candPathsMap.get(key);
	}
	
	private ArrayList<Path_2> getCandidateDPaths(
			HashMap<Integer, ArrayList<Path_2>> dPathMap,
			int pivotNode) {
		if (!dPathMap.containsKey(pivotNode))
			return new ArrayList<Path_2>();
		return dPathMap.get(pivotNode);
	}
	
	private ArrayList<Path_2> getCandidateDPathsByPairs(
			HashMap<PairNodes, ArrayList<Path_2>> dPathMap,
			int pivotNode1, int pivotNode2) {
		PairNodes pn = new PairNodes(pivotNode1, pivotNode2);
		if (!dPathMap.containsKey(pn))
			return new ArrayList<Path_2>();
		return dPathMap.get(pn);
	}
	
	private void propagateProduction(Path_1 infer) {
		if (!r1.contains(infer)) {
			r1.add(infer);
			if (IPathType_1.toBeExpanded(infer.pathType)) {
				w1.offer(infer);
				pushToIndexes(srcCandPathsMap, infer.pathType, infer.src, infer);
				pushToIndexes(destCandPathsMap, infer.pathType, infer.tgt, infer);
				int pathType = infer.pathType;
				if (IPathType_1.isRightBrace(pathType)) {
					if (!typeRBraceMap.containsKey(pathType)) 
						typeRBraceMap.put(pathType, new ArrayList<Path_1>());
					typeRBraceMap.get(pathType).add(infer);
				}
			}
		}
	}
	
	private void propagateDoubleProduction(Path_2 dp) {
		/*
		 * Date: 2014-4-24
		 * (a,b,c,d) not inserted when (a,d) appends 
		 */
		if (r1.contains(new Path_1(dp.src1, dp.tgt2, dp.pathType))) return;
		
		if (!r2.contains(dp)) {
			r2.add(dp);
			if (IPathType_2.toBeExpanded(dp.pathType)) {
				w2.offer(dp);
				// build 1D indexes
				pushToSdMap(s1Map, dp.src1, dp);
				pushToSdMap(d1Map, dp.tgt1, dp);
				pushToSdMap(s2Map, dp.src2, dp);
				pushToSdMap(d2Map, dp.tgt2, dp);
				// build 2D indexes for (dest1,src2) & (src1,dest2)
				pushToPairMap(innerPairMap, dp.tgt1, dp.src2, dp);
				pushToPairMap(outerPairMap, dp.src1, dp.tgt2, dp);
			}
		}
	}
	
	private void propagateDoubleProduction(Path_1 p1, Path_1 p2, int dPathType) {
		propagateDoubleProduction(new Path_2(p1.src, p1.tgt, p2.src, p2.tgt, dPathType));
	}
	
	/*
	 * A -> @B C 
	 */
	private void propagateRightAppend(int typeA, Path_1 pB, int typeC) {
		ArrayList<Path_1> candPaths 
			= getCandidatePaths(srcCandPathsMap, typeC, pB.tgt);
		for (Path_1 pC : candPaths) {
			Path_1 pA = new Path_1(pB.src, pC.tgt, typeA);
			propagateProduction(pA);
			//System.out.println("A -> @B C ");
		}
	}
	/*
	 * A -> B @C
	 */
	private void propagateLeftAppend(int typeA, int typeB, Path_1 pC) {
		ArrayList<Path_1> candPaths 
			= getCandidatePaths(destCandPathsMap, typeB, pC.src);
		for (Path_1 pB : candPaths) {
			Path_1 pA = new Path_1(pB.src, pC.tgt, typeA);
			propagateProduction(pA);
			//System.out.println("A -> B @C");
		}
	}	
	/*
	 * A -> @a
	 */
	private void propagateTrans(int typeA, Path_1 pa) {
		Path_1 pA = new Path_1(pa.src, pa.tgt, typeA);
		propagateProduction(pA);
		//System.out.println("A -> @a");
	}
	
	/*
	 * &S -> plus( [{i] }i )
	 */
	private void propagatePlusBraceLeft(int typeA, Path_1 leftBrace) {
		int braceType = leftBrace.pathType;
		ArrayList<Path_1> pairedRBrace = typeRBraceMap.get(-braceType);
		if (pairedRBrace != null)
			for (Path_1 rightBrace : pairedRBrace) {
				propagateDoubleProduction(leftBrace, rightBrace, IPathType_2.SS_S);
			}
	}
	
	/*
	 * Date: 2014-4-9
	 * A -> 'MINUS'(&B)
	 */
	private void propagateMinus(int typeA, Path_2 dpB) {
		if (dpB.tgt1 == dpB.src2) { 
			// src1 --> tgt1 = src2 --> tgt2
			Path_1 pS = new Path_1(dpB.src1, dpB.tgt2, typeA);
			propagateProduction(pS);
		}
	}
	
	/*
	 * Date: 2014-4-9
	 * &A -> 'OUT-LEFT'(@&B, C)
	 */
	private void propagateOuterLeft1(int typeA, Path_2 dpB, int typeC) {
		ArrayList<Path_1> candPaths 
			= getCandidatePaths(destCandPathsMap, typeC, dpB.src1);
		for (Path_1 pC : candPaths) {
			Path_2 dpA = 
					new Path_2(
							pC.src, dpB.tgt1,
							dpB.src2, dpB.tgt2, 
							typeA
					);
			propagateDoubleProduction(dpA);
		}
	}
	
	/*
	 * Date: 2014-4-9
	 * &A -> 'OUT-LEFT'(&B, @C)
	 */
	private void propagateOuterLeft2(int typeA, Path_1 pC) {
		ArrayList<Path_2> candDPaths
			= getCandidateDPaths(s1Map, pC.tgt); 
		for (Path_2 dpB : candDPaths) {
			Path_2 dpA = 
					new Path_2(
							pC.src, dpB.tgt1,
							dpB.src2, dpB.tgt2, 
							typeA
					);
			propagateDoubleProduction(dpA);
		}
	}
	
	/*
	 * Date: 2014-4-9
	 * &A -> 'IN-LEFT'(@&B, C)
	 */
	private void propagateInnerLeft1(int typeA, Path_2 dpB, int typeC) {
		ArrayList<Path_1> candPaths 
			= getCandidatePaths(srcCandPathsMap, typeC, dpB.tgt1);
		for (Path_1 pC : candPaths) {
			Path_2 dpA = 
					new Path_2(
							dpB.src1, pC.tgt,
							dpB.src2, dpB.tgt2, 
							typeA
					);
			propagateDoubleProduction(dpA);
		}
	}
	
	/*
	 * Date: 2014-4-9
	 * &A -> 'IN-LEFT'(&B, @C)
	 */
	private void propagateInnerLeft2(int typeA, Path_1 pC) {
		ArrayList<Path_2> candDPaths
			= getCandidateDPaths(d1Map, pC.src); 
		for (Path_2 dpB : candDPaths) {
			Path_2 dpA = 
					new Path_2(
							dpB.src1, pC.tgt,
							dpB.src2, dpB.tgt2, 
							typeA
					);
			propagateDoubleProduction(dpA);
		}
	}	
	
	/*
	 * Date: 2014-4-9
	 * &A -> 'IN-RIGHT'(@&B, C)
	 */
	private void propagateInnerRight1(int typeA, Path_2 dpB, int typeC) {
		ArrayList<Path_1> candPaths 
			= getCandidatePaths(destCandPathsMap, typeC, dpB.src2);
		for (Path_1 pC : candPaths) {
			Path_2 dpA = 
					new Path_2(
							dpB.src1, dpB.tgt1,
							pC.src, dpB.tgt2,
							typeA
					);
			propagateDoubleProduction(dpA);
		}
	}
	
	/*
	 * Date: 2014-4-9
	 * &A -> 'IN-RIGHT'(&B, @C)
	 */
	private void propagateInnerRight2(int typeA, Path_1 pC) {
		ArrayList<Path_2> candDPaths
			= getCandidateDPaths(s2Map, pC.tgt); 
		for (Path_2 dpB : candDPaths) {
			Path_2 dpA = 
					new Path_2(
							dpB.src1, dpB.tgt1,
							pC.src, dpB.tgt2,
							typeA
					);
			propagateDoubleProduction(dpA);
		}
	}
	
	/*
	 * Date: 2014-4-9
	 * &A -> 'OUT-RIGHT'(@&B, C)
	 */
	private void propagateOuterRight1(int typeA, Path_2 dpB, int typeC) {
		ArrayList<Path_1> candPaths 
			= getCandidatePaths(srcCandPathsMap, typeC, dpB.tgt2);
		for (Path_1 pC : candPaths) {
			Path_2 dpA = 
					new Path_2(
							dpB.src1, dpB.tgt1,
							dpB.src2, pC.tgt,
							typeA
					);
			propagateDoubleProduction(dpA);
		}
	}
	
	/*
	 * Date: 2014-4-9
	 * &A -> 'OUT-RIGHT'(&B, @C)
	 */
	private void propagateOuterRight2(int typeA, Path_1 pC) {
		ArrayList<Path_2> candDPaths
			= getCandidateDPaths(d2Map, pC.src); 
		for (Path_2 dpB : candDPaths) {
			Path_2 dpA = 
					new Path_2(
							dpB.src1, dpB.tgt1,
							dpB.src2, pC.tgt,
							typeA
					);
			propagateDoubleProduction(dpA);
		}
	}	
	
	/*
	 * Date: 2014-4-9
	 * &A -> 'DIntersec'(@Outer&B, Inner&C)
	 */
	private void propagateInnerDI(int typeA, Path_2 dpB) {
		ArrayList<Path_2> candDPaths
			= getCandidateDPathsByPairs(outerPairMap, dpB.tgt1, dpB.src2); 
		for (Path_2 dpC : candDPaths) {
			Path_2 dpA = 
					new Path_2(
							dpB.src1, dpC.tgt1,
							dpC.src2, dpB.tgt2,
							typeA
					);
			propagateDoubleProduction(dpA);
		}
	}
	
	/*
	 * Date: 2014-4-9
	 * &A -> 'DIntersec'(Outer&B, @Inner&C)
	 */
	private void propagateOuterDI(int typeA, Path_2 dpC) {
		ArrayList<Path_2> candDPaths
			= getCandidateDPathsByPairs(innerPairMap, dpC.src1, dpC.tgt2); 
		for (Path_2 dpB : candDPaths) {
			Path_2 dpA = 
					new Path_2(
							dpB.src1, dpC.tgt1,
							dpC.src2, dpB.tgt2,
							typeA
					);
			propagateDoubleProduction(dpA);
		}
	}

	public void talAlgorithm() {
		/*
		 * 'e' : labelled 0
		 * '{_i' : labelled i
		 * '}_i' : labelled -i
		 */
		/* Init the worklist */
		int n = g.getN();
		for (int srcId = 0; srcId < n; srcId++) {
			g.initializeForwardEdgeIterator(srcId);
			while (g.hasNextEdge()) {
				int tgtId = g.getForwardEdgeTarget();
				int pathType = g.getEdgeType();
				//if (tgtId < n)
					propagateProduction(new Path_1(srcId, tgtId, pathType));
			}
		}
		//System.gc();
		/* Add Reachability for production */
		// TODO (a,b,c,d) will be eliminated when (a,d) appears
		int cnt = 0;
		while (!w1.isEmpty() || !w2.isEmpty()) {
			cnt++;
			if (cnt % 2500 == 0) {
				System.out.println(cnt + " " + r1.size() + " " + w1.size() 
						+ " " + r2.size() + " " + w2.size());
				//if (cnt % 10000 == 0)
				//	System.gc(); //////
			}
			if (!w1.isEmpty()) {
				Path_1 p = w1.poll();
				switch (p.pathType) {
					case IPathType_1.S_S:
						// S -> @SS
						propagateRightAppend(IPathType_1.S_S, p, IPathType_1.S_S);
						// S -> S@S
						propagateLeftAppend(IPathType_1.S_S, IPathType_1.S_S, p);
						// [Abandoned] S -> {i@Si}
						// propagateBraces(p);
						// L -> L@S
						propagateLeftAppend(IPathType_1.S_SL, IPathType_1.S_SL, p);
						// L -> @SL
						propagateRightAppend(IPathType_1.S_SL, p, IPathType_1.S_SL);
						// R -> R@S
						propagateLeftAppend(IPathType_1.S_SR, IPathType_1.S_SR, p);
						// R -> @SR
						propagateRightAppend(IPathType_1.S_SR, p, IPathType_1.S_SR);
						// &S -> 'OL'(&S, @S)
						propagateOuterLeft2(IPathType_2.SS_S, p);
						// &S -> 'IL'(&S, @S)
						propagateInnerLeft2(IPathType_2.SS_S, p);
						// &S -> 'IR'(&S, @S)
						propagateInnerRight2(IPathType_2.SS_S, p);
						// &S -> 'OR'(&S, @S)
						propagateOuterRight2(IPathType_2.SS_S, p);
						break;
					case IPathType_1.S_SL:
						// L -> @LS
						propagateRightAppend(IPathType_1.S_SL, p, IPathType_1.S_S);
						// L -> S@L
						propagateLeftAppend(IPathType_1.S_SL, IPathType_1.S_S, p);
						// L -> @LL
						propagateRightAppend(IPathType_1.S_SL, p, IPathType_1.S_SL);
						// L -> L@L
						propagateLeftAppend(IPathType_1.S_SL, IPathType_1.S_SL, p);
						// M -> R@L
						propagateLeftAppend(IPathType_1.S_SM, IPathType_1.S_SR, p);
						// &L -> 'OL'(&S, @L)
						propagateOuterLeft2(IPathType_2.SS_SL, p);
						// &L -> 'OR'(&S, @L)
						propagateOuterRight2(IPathType_2.SS_SL, p);
						break;
					case IPathType_1.S_SR:
						// R -> @RS
						propagateRightAppend(IPathType_1.S_SR, p, IPathType_1.S_S);
						// R -> S@R
						propagateLeftAppend(IPathType_1.S_SR, IPathType_1.S_S, p);
						// R -> @RR
						propagateRightAppend(IPathType_1.S_SR, p, IPathType_1.S_SR);
						// R -> R@R
						propagateLeftAppend(IPathType_1.S_SR, IPathType_1.S_SR, p);
						// M -> @RL
						propagateRightAppend(IPathType_1.S_SM, p, IPathType_1.S_SL);
						// &R -> 'OL'(&S, R)
						propagateOuterLeft2(IPathType_2.SS_SR, p);
						// &R -> 'OR'(&S, R)
						propagateOuterRight2(IPathType_2.SS_SR, p);
						break;
					case IPathType_1.S_SM: 
						// will not be processed 
						// since it does not appear in the right-side of any production
						break;
					default:
						if (p.pathType > 0) {
						// L -> @{i\
							propagateTrans(IPathType_1.S_SL, p);
						// 'PLUS' @{i }i
							propagatePlusBraceLeft(IPathType_1.S_S, p);
						}
						else {
						// R -> @}i
							propagateTrans(IPathType_1.S_SR, p);
						// 'PLUS' {i @}i  (can be processed in 'propagatePlusBraceLeft')
							//propagatePlusBraceRight(IPathType_1.S_S, -p.pathType, p);
						}
					}
			}
			else if (!w2.isEmpty()) {
				/*
				 * Date: 2014-4-9
				 */
				Path_2 dp = w2.poll();
				/*
				 * Date: 2014-4-24
				 * remove (a,b,c,d) when (a,d) happens
				 */
				if (r1.contains(new Path_1(dp.src1, dp.tgt2, dp.pathType))) {
					continue;
				}
				switch (dp.pathType) {
					case IPathType_2.SS_S:
						// S -> 'MINUS'(@&S)
						propagateMinus(IPathType_1.S_S, dp);
						// &S -> 'OL'(@&S, S)
						propagateOuterLeft1(IPathType_2.SS_S, dp, IPathType_1.S_S);
						// &S -> 'IL'(@&S, S)
						propagateInnerLeft1(IPathType_2.SS_S, dp, IPathType_1.S_S);
						// &S -> 'IR'(@&S, S)
						propagateInnerRight1(IPathType_2.SS_S, dp, IPathType_1.S_S);
						// &S -> 'OR'(@&S, S)
						propagateOuterRight1(IPathType_2.SS_S, dp, IPathType_1.S_S);						
						// &S -> 'DIntersec'( @Outer&S, Inner&S)
						propagateInnerDI(IPathType_2.SS_S, dp);						
						// &S -> 'DIntersec'( Outer&S, @Inner&S)
						propagateOuterDI(IPathType_2.SS_S, dp);
						// Date: 2014-5-14
						// Debug: 
						// &L -> 'OL'(@&S, L)
						propagateOuterLeft1(IPathType_2.SS_SL, dp, IPathType_1.S_SL);
						// &L -> 'OR'(@&S, L)
						propagateOuterRight1(IPathType_2.SS_SL, dp, IPathType_1.S_SL);
						// &R -> 'OL'(&S, R)
						propagateOuterLeft1(IPathType_2.SS_SR, dp, IPathType_1.S_SR);
						// &R -> 'OR'(&S, R)
						propagateOuterRight1(IPathType_2.SS_SR, dp, IPathType_1.S_SR);
						break;
					case IPathType_2.SS_SL:
						// L -> 'MINUS'(@&SL)
						propagateMinus(IPathType_1.S_SL, dp);
						break;
					case IPathType_2.SS_SR:
						// SR -> 'MINUS'(@&SR)
						propagateMinus(IPathType_1.S_SR, dp);
						break;
				}
			}
		}	
		System.out.println("|r1| = " + r1.size());
		System.out.println("|r2| = " + r2.size());
		//int r2minusr1 = 0;
		
		HashSet<Path_2> hashSetPath2 = new HashSet<Path_2>();
		Iterator<Path_2> p2_it = r2.iterator();
		while (p2_it.hasNext()) {
			Path_2 dp = p2_it.next();
			if (g.getNodeType(dp.src1) != INodeType.NORMAL_NODE &&
					g.getNodeType(dp.tgt1) != INodeType.NORMAL_NODE &&
					g.getNodeType(dp.src2) != INodeType.NORMAL_NODE &&
					g.getNodeType(dp.tgt2) != INodeType.NORMAL_NODE) {
				if (!r1.contains(new Path_1(dp.src1, dp.tgt2, dp.pathType))) {
					//r2minusr1 ++;
					hashSetPath2.add(dp);
				}
			}
		}
		System.out.println("|r2| = " + hashSetPath2.size());
	}

}
